@@ -1,14 +1,17 @@
 from unittest import TestCase
 from dogpile.cache.api import NO_VALUE, CachedValue
+from dogpile.cache.api import CacheBackend
 from dogpile.cache import exception
 from dogpile.cache import make_region, CacheRegion
 from dogpile.cache.proxy import ProxyBackend
 from dogpile.cache.region import _backend_loader
 from dogpile.cache.region import RegionInvalidationStrategy
+from dogpile.util import compat
 from . import eq_, is_, assert_raises_message, io, configparser
 import time
 import datetime
 import itertools
+import mock
 from collections import defaultdict
 from ._fixtures import MockBackend
 
@@ -17,6 +20,11 @@ def key_mangler(key):
     return "HI!" + key
 
 
+class APITest(TestCase):
+    def test_no_value_str(self):
+        eq_(str(NO_VALUE), "<dogpile.cache.api.NoValue object>")
+
+
 class RegionTest(TestCase):
 
     def _region(self, init_args={}, config_args={}, backend="mock"):
@@ -51,7 +59,7 @@ class RegionTest(TestCase):
 
         my_region = make_region()
         config = configparser.ConfigParser()
-        config.readfp(io.StringIO(my_conf))
+        compat.read_config_file(config, io.StringIO(my_conf))
 
         my_region.configure_from_config(
             dict(config.items('xyz')), 'cache.example.')
@@ -147,6 +155,15 @@ class RegionTest(TestCase):
         )
         eq_(reg.is_configured, False)
 
+    def test_invalid_backend(self):
+        reg = CacheRegion()
+        assert_raises_message(
+            exception.PluginNotFound,
+            "Couldn't find cache plugin to load: unknown",
+            reg.configure, 'unknown'
+        )
+        eq_(reg.is_configured, False)
+
     def test_set_get_value(self):
         reg = self._region()
         reg.set("some key", "some value")
@@ -270,6 +287,41 @@ class RegionTest(TestCase):
         eq_(reg.get_or_create("some key", creator),
             "some value 2")
 
+        eq_(reg.get_or_create("some key", creator),
+            "some value 2")
+
+        reg.invalidate()
+        eq_(reg.get_or_create("some key", creator),
+            "some value 3")
+
+        eq_(reg.get_or_create("some key", creator),
+            "some value 3")
+
+    def test_hard_invalidate_get_or_create_multi(self):
+        reg = self._region()
+        counter = itertools.count(1)
+
+        def creator(*keys):
+            return ["some value %s %d" % (k, next(counter)) for k in keys]
+
+        eq_(reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 1", "some value k2 2"])
+
+        time.sleep(.1)
+        reg.invalidate()
+        eq_(reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 3", "some value k2 4"])
+
+        eq_(reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 3", "some value k2 4"])
+
+        reg.invalidate()
+        eq_(reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 5", "some value k2 6"])
+
+        eq_(reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 5", "some value k2 6"])
+
     def test_soft_invalidate_get(self):
         reg = self._region(config_args={"expiration_time": 1})
         reg.set("some key", "some value")
@@ -477,6 +529,79 @@ class CustomInvalidationStrategyTest(RegionTest):
         return reg
 
 
+class AsyncCreatorTest(TestCase):
+    def _fixture(self):
+
+        def async_creation_runner(cache, somekey, creator, mutex):
+            try:
+                value = creator()
+                cache.set(somekey, value)
+            finally:
+                mutex.release()
+
+        return mock.Mock(side_effect=async_creation_runner)
+
+    def test_get_or_create(self):
+        acr = self._fixture()
+        reg = CacheRegion(async_creation_runner=acr)
+        reg.configure("mock", expiration_time=.2)
+
+        def some_value():
+            return "some value"
+
+        def some_new_value():
+            return "some new value"
+
+        eq_(reg.get_or_create("some key", some_value), "some value")
+        time.sleep(.5)
+        eq_(
+            reg.get_or_create("some key", some_new_value),
+            "some value")
+        eq_(
+            reg.get_or_create("some key", some_new_value),
+            "some new value")
+        eq_(
+            acr.mock_calls, [
+                mock.call(reg, "some key",
+                          some_new_value, reg._mutex("some key"))
+            ])
+
+    def test_fn_decorator(self):
+        acr = self._fixture()
+        reg = CacheRegion(async_creation_runner=acr)
+        reg.configure("mock", expiration_time=5)
+
+        canary = mock.Mock()
+
+        @reg.cache_on_arguments()
+        def go(x, y):
+            canary(x, y)
+            return x + y
+
+        eq_(go(1, 2), 3)
+        eq_(go(1, 2), 3)
+
+        eq_(canary.mock_calls, [mock.call(1, 2)])
+
+        eq_(go(3, 4), 7)
+
+        eq_(canary.mock_calls, [mock.call(1, 2), mock.call(3, 4)])
+
+        reg.invalidate(hard=False)
+
+        eq_(go(1, 2), 3)
+
+        eq_(canary.mock_calls, [
+            mock.call(1, 2), mock.call(3, 4), mock.call(1, 2)])
+
+        eq_(
+            acr.mock_calls, [
+                mock.call(reg, "tests.cache.test_region:go|1 2",
+                          mock.ANY,
+                          reg._mutex("tests.cache.test_region:go|1 2"))
+            ])
+
+
 class ProxyBackendTest(TestCase):
 
     class GetCounterProxy(ProxyBackend):
@@ -613,3 +738,19 @@ class ProxyBackendTest(TestCase):
         # make sure 1 was set, but 5 was not
         eq_(reg.get(5), NO_VALUE)
         eq_(reg.get(1), True)
+
+    def test_actual_backend_proxied(self):
+        # ensure that `reg.actual_backend` is the actual backend
+        # also ensure that `reg.backend` is a proxied backend
+        reg = self._region(config_args={"wrap": [
+            ProxyBackendTest.GetCounterProxy,
+            ProxyBackendTest.SetCounterProxy]})
+        assert isinstance(reg.backend, ProxyBackend)
+        assert isinstance(reg.actual_backend, CacheBackend)
+
+    def test_actual_backend_noproxy(self):
+        # ensure that `reg.actual_backend` is the actual backend
+        # also ensure that `reg.backend` is NOT a proxied backend
+        reg = self._region()
+        assert isinstance(reg.backend, CacheBackend)
+        assert isinstance(reg.actual_backend, CacheBackend)
